library ormicida.benchmarks._benchs;

import 'package:benchmark_harness/benchmark_harness.dart';
import 'dart:convert' as convert;
import 'package:bson/bson.dart';
import 'package:ormicida_bson/ormicida_bson.dart';
import 'package:ormicida_codec/ormicida_codec.dart';
import 'package:ormicida_json/ormicida_json.dart';
import 'package:ormicida_pixm/ormicida_pixm.dart';
import 'package:ormicida_schema/ormicida_schema.dart';
import 'package:ormicida_schema_builder/ormicida_schema_builder.dart';
import 'package:ormicida_schema_inspector/ormicida_schema_inspector.dart';

export 'package:ormicida_bson/ormicida_bson.dart' show bson;
export 'package:ormicida_json/ormicida_json.dart' show json;
export 'package:ormicida_pixm/ormicida_pixm.dart' show pixm;

class _Score {
  final String  name;
  final double  time;
  double        ratio;
  
  _Score(this.name, this.time);
  
  void _pad(StringBuffer buffer, int code, int length) {
    while (length-- > 0) buffer.writeCharCode(code);
  }
  void _padLeft(StringBuffer buffer, String label, int length) {
    _pad(buffer, 32, length - label.length);
    buffer.write(label);
  }
  void _padRight(StringBuffer buffer, String label, int length) {
    buffer.write(label);
    _pad(buffer, 32, length - label.length);
  }
  
  void _printValue(double value, int maxLen) {
    final buffer = new StringBuffer();
    buffer.write('  ');
    _padRight(buffer, name, maxLen + 2);
    _padLeft(buffer, value.toStringAsFixed(2), 12);
    print(buffer.toString());
  }
  
  void printTime(int maxLen) {
    _printValue(time, maxLen);
  }
  
  void printRatio(int maxLen) {
    _printValue(ratio, maxLen);
  }
}

class BenchmarkSet {
  final String name;
  final _benchs = <BenchmarkBase>[];
  
  BenchmarkSet(this.name);
  
  void add(BenchmarkBase bench) {
    _benchs.add(bench);
  }
  
  void report({bool sort: true}) {
    print('###########  $name');
    
    final maxLen = _benchs.fold(0,
        (int current, BenchmarkBase bench) {
          final length = bench.name.length;
          return length > current ? length : current;
        });
    final scores = <_Score>[];
    double minTime = double.MAX_FINITE;
    
    _benchs.forEach((bench) {
      final time = bench.measure();
      final score = new _Score(bench.name, time);
      
      if (time < minTime) minTime = time;
      scores.add(score);
      
      score.printTime(maxLen);
    });
    
    print('');
    
    if (sort) {
      scores.sort((score1, score2) => score1.time.compareTo(score2.time));
    }
    
    scores.forEach((score) {
      score.ratio = score.time / minTime;
      score.printRatio(maxLen);
    });
    
    print('');
  }
}

int _counter = 0;

class SchemaEncodeBench extends BenchmarkBase {
  final CodecEncoder encoder;
  final object;
  final Schema schema;
  List<int> result;
  
  SchemaEncodeBench(String name, this.encoder, this.object, this.schema): super(name);
  
  void run() {
    result = inspector.encode(object, encoder, schema);
    if (result != null) _counter = 0;
  }
}

class SchemaDecodeBench extends BenchmarkBase {
  final CodecDecoder decoder;
  final List<int> bytes;
  final Schema schema;
  final bool forceDynamic;
  Object result;
  
  SchemaDecodeBench(String name, this.decoder, this.bytes, this.schema, this.forceDynamic): super(name);
  
  void run() {
    result = builder.decode(bytes, decoder, schema, forceDynamic: forceDynamic);
    if (result != null) _counter = 1;
  }
}

class DynamicEncodeBench extends BenchmarkBase {
  final _DynamicCodec  codec;
  final               object;
  List<int>           result;
  DynamicEncodeBench(String name, this.codec, this.object): super(name);
  void run() {
    result = codec.encode(object);
    if (result != null) _counter = 0;
  }
}

class DynamicDecodeBench extends BenchmarkBase {
  final _DynamicCodec codec;
  final List<int> bytes;
  dynamic result;
  DynamicDecodeBench(String name, this.codec, this.bytes): super(name);
  void run() {
    result = codec.decode(bytes);
    if (result != null) _counter = 1;
  }
}

class DynamicTotalBench extends BenchmarkBase {
  final _DynamicCodec codec;
  final               object;
  List<int>           bytes;
  var                 result;
  DynamicTotalBench(String name, this.codec, this.object): super(name);
  void run() {
    bytes = codec.encode(object);
    result = codec.decode(bytes);
    if (result != null) _counter = 2;
  }
}

abstract class _DynamicCodec {
  List<int> encode(object);
  dynamic decode(List<int> bytes);
}

const CodecJson CODEC_JSON = const CodecJson();
class CodecJson implements _DynamicCodec {
  const CodecJson();
  List<int> encode(object) => json.encode(object);
  dynamic decode(List<int> bytes) => json.decode(bytes);
}

const CodecPixm CODEC_PIXM = const CodecPixm();
class CodecPixm implements _DynamicCodec {
  const CodecPixm();
  List<int> encode(object) => pixm.encode(object);
  dynamic decode(List<int> bytes) => pixm.decode(bytes);
}

const CodecBson CODEC_BSON = const CodecBson();
class CodecBson implements _DynamicCodec {
  const CodecBson();
  List<int> encode(object) => bson.encode(object);
  dynamic decode(List<int> bytes) => bson.decode(bytes);
}

const CodecConvert CODEC_CONVERT = const CodecConvert();
class CodecConvert implements _DynamicCodec {
  const CodecConvert();
  List<int> encode(object) =>
      convert.UTF8.encode(convert.JSON.encode(object));
  dynamic decode(List<int> bytes) =>
      convert.JSON.decode(convert.UTF8.decode(bytes));
}

final _bson2 = new BSON();
const CodecBson2 CODEC_BSON2 = const CodecBson2();
class CodecBson2 implements _DynamicCodec {
  const CodecBson2();
  List<int> encode(object) =>
      _bson2.serialize(object).byteList;
  dynamic decode(List<int> bytes) =>
      _bson2.deserialize(new BsonBinary.from(bytes));
}
